import os
import sys
import types
import inspect


def virtual(spec, name):

    """
    Create and insert a fake module. Although this *could* be imported, its
    intended use is providing a means of grouping sub-modules in an independent
    namespace. Only the respective qual-name is returned instead of the module.
    """

    qual = spec.name + '.' + name

    module = types.ModuleType(qual)

    module.__spec__ = spec

    marker = object()

    module.__path__ = marker

    def find_spec(name, path, target = None):

        return module.__spec__ if path is marker else None

    finder = types.SimpleNamespace(find_spec = find_spec)

    sys.meta_path.append(finder)

    return qual


def parental(path):

    """
    Get the main enterance point path denoting packages.
    """

    return os.path.join(path, '__init__.py') if os.path.isdir(path) else path


def filial(module, name):

    """
    Get the path of a submodule file of the parent.
    """

    path = next(iter(module.__path__), module.__file__)

    (directory, junk) = os.path.split(path)

    return os.path.join(directory, name)


__result = object()


def findcall(source, name, args = (), kwargs = {}, result = __result):

    try:

        function = getattr(source, name)

    except KeyError:

        return result

    value = function(*args, **kwargs)

    def final():

        return value if result is __result else result

    if inspect.iscoroutine(value):

        async def function():

            nonlocal value

            value = await value

            return final()

        return function()

    return final()



import os
import sys
import types
import itertools
import importlib
import collections
import weakref
import inspect

from . import helpers


__all__ = ('Client', 'load', 'drop', 'origin')


_cache = weakref.WeakValueDictionary()


def _register(unique, module):

    try:

        store = _cache[unique]

    except KeyError:

        store = _cache[unique] = weakref.WeakValueDictionary()

    store[module.__name__] = module


class Client:

    """
    Main means of interacting with the segmentation API.

    Upon creation, a fake module will be created and inserted into
    ``sys.modules`` with a unique numerical identifier appended to it. This is
    the base for loading any subsequent modules.
    """

    __slots__ = ('_modules', '_name')

    _uniques = set()

    _lead = helpers.virtual(__spec__, '_')

    def __init__(self):

        self._modules = {}

        indexes = itertools.count(0)

        identities = map(str, indexes)

        check = self._uniques.__contains__

        generate = itertools.filterfalse(check, identities)

        unique = next(generate)

        self._uniques.add(unique)

        self._name = f'{self._lead}.{unique}'

    @property
    def modules(self):

        return self._modules

    def _load(self, module):

        pass

    def load(self, path):

        """
        Create and track a new module.

        .. note::

            Must be called directly at least once before the global
            :func:`load` can be used.
        """

        base = os.path.basename(path)

        (name, extension) = os.path.splitext(base)

        path = helpers.parental(path)

        name = f'{self._name}.{name}'

        spec = importlib.util.spec_from_file_location(name, path)

        module = importlib.util.module_from_spec(spec)

        self._load(module)

        self._modules[path] = module

        _register(self, module)

        sys.modules[name] = module

        module.__spec__.loader.exec_module(module)

        return module

    def _drop(self, module):

        pass

    def drop(self, path):

        path = helpers.parental(path)

        module = self._modules.pop(path)

        self._drop(module)

        for key in tuple(sys.modules):

            if not key.startswith(module.__name__):

                continue

            sys.modules[key]

        return module

    def __del__(self):

        parts = self._name.rsplit('.', 1)

        unique = parts[-1]

        self._uniques.remove(unique)


def _origin():

    for info in inspect.stack():

        space = info.frame.f_globals

        name = space['__name__']

        for (unique, modules) in _cache.items():

            try:

                module = modules[name]

            except KeyError:

                continue

            if not space is module.__dict__:

                continue

            break

        else:

            continue

        break

    else:

        raise ModuleNotFoundError('Source is not an internally loaded module.')

    return (unique, module)


def load(path, *args, **kwargs):

    """
    Find the client for the source module and load the path.

    .. warning::

        :meth:`Client.load` must be called at least once before this is used.
    """

    (client, module) = _origin()

    path = helpers.filial(module, path)

    module = client.load(path)

    name = inspect.stack()[0].function

    result = helpers.findcall(module, name, args = args, kwargs = kwargs)

    return result


def drop(path, *args, **kwargs):

    """
    Find the client for the source module and drop the path.
    """

    (client, module) = _origin()

    path = helpers.filial(module, path)

    client.drop(path)

    name = inspect.stack()[0].function

    result = helpers.findcall(module, name, args = args, kwargs = kwargs)

    return result


def origin(full = False):

    """
    Get the client holding the loaded module this is called from.
    """

    assets = _origin()

    return assets if full else assets[0]




"""
This extension utilizes :mod:`wrapio` to accomodate for event-based applications
as well as :mod:`lineopt` for creating command parsing and invoking frameworks.
"""

import asyncio
import wrapio
import lineopt
import collections

import incise


__all__ = ('Client', 'load', 'drop', 'call', 'wait', 'sub', 'origin')


_Asset = collections.namedtuple('Asset', 'track names')


class Client(incise.Client):

    """
    See superclass documentation for more information.

    :param asyncio.AbstractEventLoop loop:
        Used for :class:`wrapio.Track` creation.

    :var lineopt.State state:
        The universal commands mapping.
    """

    __slots__ = ('_state', '_assets', '_loop')

    def __init__(self, loop = None):

        super().__init__()

        self._state = lineopt.State()

        self._assets = {}

        self._loop = loop or asyncio.get_event_loop()

    @property
    def state(self):

        return self._state

    def invoke(self, name, *values):

        """
        Call the respective method for all tracks.
        """

        for asset in self._assets.values():

            asset.track.invoke(name, *values)

    def _load(self, module):

        track = wrapio.Track(loop = self._loop)

        self._assets[module] = _Asset(track, {})

    def _drop(self, module):

        asset = self._assets.pop(module)

        for name in asset.flags:

            del self._state[name]


load = incise.load


drop = incise.drop


def _origin():

    (client, module) = wrapio.origin(True)

    asset = client.assets[module]

    return (client, asset)


def origin(full = False):

    """
    Get the client for the source module.
    """

    values = _origin()

    return values if full else values[0]


def call(name):

    """
    Works the same as :meth:`wrapio.Track.call`.

    .. code-block::

        @call('message create')
        async def listen(event):
            print(event.message.author.username, 'said', event.message.content)
    """

    (client, asset) = _origin()

    return asset.track.call(name)


def wait(name):

    """
    Works the same as :meth:`wrapio.Track.wait`.

    .. code-block::

        @wait('message create')
        async def waiter_0(event):
            if not event.message.content == 'agreed':
                return
            waiter_0.set()
        await waiter_0.wait()
        print('agreed to the rules')
    """

    (client, asset) = _origin()

    return asset.track.wait(name)


def sub(name, *args, **kwargs):

    """
    Works the same as :meth:`lineopt.State.sub`.

    .. code-block::

        flags = {'-u': User}

        @sub('pat', flags)
        async def command_0(event, context):
            user = context.arguments['-u']
            response = f'*pat {user.username}*'
            await client.create_message(event.channel.id, content = response)

        @command_0.sub('hard')
        def comand_0_0(event, context):
            # ...
    """

    (client, asset) = _origin()

    asset.flags[name] = flags

    return client.state.sub(name, *args, **kwargs)



import setuptools

with open('README.rst') as file:

    readme = file.read()

name = 'incise'

version = '0.2.1'

author = 'Exahilosys'

url = f'https://github.com/{author}/{name}'

extensions = {
    'events': [
        'wrapio',
        'lineopt'
    ]
}

setuptools.setup(
    name = name,
    version = version,
    url = url,
    packages = setuptools.find_packages(),
    license = 'MIT',
    description = 'Live Segmentation Framework.',
    long_description = readme,
    extras_require = {
        'docs': [
            'sphinx',
            *map((name + '[{}]').format, extensions)
        ],
        **extensions
    }
)
